#include <avr/io.h>
#include <util/delay.h>
#include "common.h"
#include "extern.h"
#include "i2c.h"
#include "xitoa.h"

volatile uint16_t ms_c1, ms_c2, ms_c3, ms_c4, ms_c5, ms_c6; // calibration data from prom.
volatile uint32_t ms_d1, ms_d2;
volatile uint8_t MS5607_mode = 0;

// TWI mode
typedef enum _tagTWIMode{
	TWI_MODE_SEND,		// send
	TWI_MODE_RECEIVE,	// recv
}TWIMode;

// TWI ack
typedef enum _tagTWIACKMode{
	TWI_ACKMODE_ACK,	// ack
	TWI_ACKMODE_NACK,	// nack
}TWIACKMode;

// SCL clock = CPU_CLOCK / (16+2*(TWBR)*prescaler)
// TWBR = (CPU_CLOCK/SCL clock-16)/prescaler/2
void init_i2c(void){
	TWSR = 0; // 1/1
	//TWSR = _BV(TWPS0); // 1/4
	//TWSR = _BV(TWPS1); // 1/16
	TWSR = _BV(TWPS0) | _BV(TWPS1); // 1/64
	TWBR = 2;
	TWCR = _BV(TWEN);
}

int8_t i2c_start(void){
	TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN);

	loop_until_bit_is_set(TWCR, TWINT);

	if((TWSR & 0xF8) == 0x08){
		return 0;
	}else if((TWSR & 0xF8) == 0x10){
		return 1;
	}

	return -1;
/*
	// only I2C master
	i = 0;
	do{
		i++;
		if(i > 100){return -1;}
	}while(((TWSR & 0xF8) == 0x08) || ((TWSR & 0xF8) == 0x10));

	return 1;
*/
}

int8_t i2c_setaddr(uint8_t addr, TWIMode mode){
	TWDR = addr;
	TWCR = _BV(TWINT) | _BV(TWEN);

	loop_until_bit_is_set(TWCR, TWINT);

	if(mode == TWI_MODE_SEND){
		if((TWSR & 0xF8) == 0x18){
			return 0;
		}else if((TWSR & 0xF8) == 0x20){
			return 1;
		}
	}else if(mode == TWI_MODE_RECEIVE){
		if((TWSR & 0xF8) == 0x40){
			return 0;
		} else if((TWSR & 0xF8) == 0x48){
			return 1;
		}
	}else{
		return -1;
	}

	return -1;
}

int8_t i2c_snddata(uint8_t data){
	TWDR = data;
	TWCR = _BV(TWINT) | _BV(TWEN);

	loop_until_bit_is_set(TWCR, TWINT);

	if((TWSR & 0xF8) == 0x28){
		return 0;
	}else if((TWSR & 0xF8) == 0x30){
		return 1;
	}
	return -1;
}

int8_t i2c_rcvdata(uint8_t *data, TWIACKMode mode){
	if(TWI_ACKMODE_ACK == mode){
		TWCR = _BV(TWINT) | _BV(TWEA) | _BV(TWEN);
	}else{
		TWCR = _BV(TWINT) | _BV(TWEN);
	}

	loop_until_bit_is_set(TWCR, TWINT);

	if((TWSR & 0xF8) == 0x50){
		*data = TWDR;
		return 0;
	}else if((TWSR & 0xF8) == 0x58){
		*data = TWDR;
		return 1;
	}
	return -1;
}

void i2c_stop(void){
	TWCR = _BV(TWINT) | _BV(TWSTO) | _BV(TWEN);
}

uint8_t i2c_setdata(uint8_t id, uint8_t start, uint8_t *data, uint8_t num){
	uint8_t	i = 0;

	while(1){
		if(i2c_start() < 0){
			continue;
		}
		if(i2c_setaddr(id, TWI_MODE_SEND) != 0){
			continue;
		}
		if(i2c_snddata(start) != 0){
			continue;
		}
		for(i=0; i<num; i++){
			if(i2c_snddata(data[i]) < 0) break;
		}
		i2c_stop();
		break;
	}
	return 0;
}

uint8_t i2c_getdata(uint8_t id, uint8_t start, uint8_t *data, uint8_t num){
	uint8_t	i = 0;

	while(1){
		if(i2c_start() < 0){
			continue;
		}
		if(i2c_setaddr(id, TWI_MODE_SEND) != 0){
			continue;
		}
		if(i2c_snddata(start) != 0){
			continue;
		}
		if(i2c_start() < 0){
			continue;
		}
		if(i2c_setaddr(id + 1, TWI_MODE_RECEIVE) != 0){
			continue;
		}
		for(i=0; i<num-1; i++){
			if(i2c_rcvdata(&data[i], TWI_ACKMODE_ACK) < 0) break;
		}
		if(num - 1 != i){
			continue;
		}
		if(i2c_rcvdata(&data[i], TWI_ACKMODE_NACK) < 0){
			continue;
		}
		i2c_stop();
		break;
	}
	return 0;
}

uint8_t reset_MS5607(void){
	uint8_t prom[2];

	init_i2c();

	i2c_start();
	i2c_setaddr(MS5607_ADDR, TWI_MODE_SEND);
	i2c_snddata(0x1E); // reset
	i2c_stop();
	wait_ms(200);

	// read calibration data from prom. cannot read continuously
	i2c_getdata(MS5607_ADDR, 0xA2, prom, 2);
	ms_c1 = (((uint16_t)prom[0]) << 8) + ((uint16_t)prom[1]);
	i2c_getdata(MS5607_ADDR, 0xA4, prom, 2);
	ms_c2 = (((uint16_t)prom[0]) << 8) + ((uint16_t)prom[1]);
	i2c_getdata(MS5607_ADDR, 0xA6, prom, 2);
	ms_c3 = (((uint16_t)prom[0]) << 8) + ((uint16_t)prom[1]);
	i2c_getdata(MS5607_ADDR, 0xA8, prom, 2);
	ms_c4 = (((uint16_t)prom[0]) << 8) + ((uint16_t)prom[1]);
	i2c_getdata(MS5607_ADDR, 0xAA, prom, 2);
	ms_c5 = (((uint16_t)prom[0]) << 8) + ((uint16_t)prom[1]);
	i2c_getdata(MS5607_ADDR, 0xAC, prom, 2);
	ms_c6 = (((uint16_t)prom[0]) << 8) + ((uint16_t)prom[1]);
	/*
	// debug
	c1 = 46372;
	c2 = 43981;
	c3 = 29059;
	c4 = 27842;
	c5 = 31553;
	c6 = 28165;
	*/
	return 1;
}

// raw pressure read order
void MS5607_raw_pressure_order(void){
	// read d1
	i2c_start();
	i2c_setaddr(MS5607_ADDR, TWI_MODE_SEND);
	i2c_snddata(0x48); // OSR=4906, D1
	i2c_stop();
}

// read raw pressure after 10ms
void MS5607_raw_pressure_read(void){
	uint8_t adc_d1[3];
	//uint32_t d1;

	i2c_getdata(MS5607_ADDR, 0x00, adc_d1, 3);
	ms_d1 = (((uint32_t)adc_d1[0]) << 16) + (((uint32_t)adc_d1[1]) << 8) + ((uint32_t)adc_d1[2]);
}

// raw temp read order
void MS5607_raw_temp_order(void){
	// read d2
	i2c_start();
	i2c_setaddr(MS5607_ADDR, TWI_MODE_SEND);
	i2c_snddata(0x58); // OSR=4906, D2
	i2c_stop();
}

// read raw temp after 10ms
void MS5607_raw_temp_read(void){
	uint8_t adc_d2[3];
	//uint32_t d2;

	i2c_getdata(MS5607_ADDR, 0x00, adc_d2, 3);
	ms_d2 = (((uint32_t)adc_d2[0]) << 16) + (((uint32_t)adc_d2[1]) << 8) + ((uint32_t)adc_d2[2]);
}

//void MS5607_calc(uint32_t d1, uint32_t d2){
void MS5607_calc(void){
	// true data
	int64_t temp, p;
	// other value
	int32_t dT;
	int64_t off, sens;
	// for offset
	int64_t temp2;
	int64_t off2, sens2;
	// print
	uint8_t str[9];
	
	// calc temp
	dT = (int32_t)(ms_d2) - ((int32_t)(ms_c5) << 8);
	temp = (int64_t)(2000) + (((int64_t)(dT) * (int64_t)(ms_c6)) >> 23);
	if(temp > 2000){
		temp2 = 0;
		off2 = 0;
		sens2 = 0;
	}else if(temp <= 2000 && temp > -1500){
		temp2 = ((int64_t)(dT) * (int64_t)(dT)) >> 31;
		off2 = ((61 * ((int64_t)(temp) - 2000) * ((int64_t)(temp) - 2000)) >> 4);
		sens2 = 2 * ((int64_t)(temp) - 2000) * ((int64_t)(temp) - 2000);
	}else if(temp <= -1500){
		temp2= ((int64_t)(dT) * (int64_t)(dT)) >> 31;
		off2 = 61 * (((int64_t)(temp) - 2000) * ((int64_t)(temp) - 2000) >> 4) + 15 * ((int64_t)(temp) + 1500) * ((int64_t)(temp) + 1500);
		sens2 = 2 * ((int64_t)(temp) - 2000) * ((int64_t)(temp) - 2000) + 8 * ((int64_t)(temp) + 1500) * ((int64_t)(temp) + 1500);
	}else{
		temp2 = 0;
		off2 = 0;
		sens2 = 0;
	}

	// calc pressure
	temp -= temp2;
	//temp -= 450;
	off = (((int64_t)ms_c2) << 17) + (((int64_t)(ms_c4) * (int64_t)(dT)) >> 6);
	off -= off2;
	sens = (((int64_t)ms_c1) << 16) + (((int64_t)(ms_c3) * (int64_t)(dT)) >> 7);
	sens -= sens2;
	p = ((((int64_t)(ms_d1) * sens) >> 21) - off) >> 15;

	temparature = temp;
	pressure = p;

	//temp = 2000;
	xsprintf((char *)str, PSTR("%08lu"), temp);
	digit_temparature[0] = str[0] - 48;
	digit_temparature[1] = str[1] - 48;
	digit_temparature[2] = str[2] - 48;
	digit_temparature[3] = str[3] - 48;
	digit_temparature[4] = str[4] - 48;
	digit_temparature[5] = str[5] - 48;
	digit_temparature[6] = str[6] - 48;
	digit_temparature[7] = str[7] - 48;

	//p = 123467;
	xsprintf((char *)str, PSTR("%08lu"), p);
	digit_pressure[0] = str[0] - 48;
	digit_pressure[1] = str[1] - 48;
	digit_pressure[2] = str[2] - 48;
	digit_pressure[3] = str[3] - 48;
	digit_pressure[4] = str[4] - 48;
	digit_pressure[5] = str[5] - 48;
	digit_pressure[6] = str[6] - 48;
	digit_pressure[7] = str[7] - 48;
}

// MS5607
void MS5607_get(void){
	switch(MS5607_mode){
		case 0:
			MS5607_raw_pressure_order();
			MS5607_mode++;
			break;

		case 1:
			MS5607_raw_pressure_read();
			MS5607_mode++;
			break;

		case 2:
			MS5607_raw_temp_order();
			MS5607_mode++;
			break;

		case 3:
			MS5607_raw_temp_read();
			MS5607_mode++;
			break;

		case 4:
			MS5607_calc();
			MS5607_mode++;
			break;

		default:
			MS5607_mode = 0;
			break;
	}
}
